//	on-the-fly c and asm code colorizer v1.00, by Morgwn McCarty on 5/21/94

//	instructions:
//		load "coloriz.b" by using "load-bytes"
//		save state file with "save-state"
//		exit Epsilon and re-enter
//		note: replaces several functions, so you might want to take a look
//			at the "REPLACE_FUNC" commands near the end of this file

//	improvements over previous comment highlighters:
//		no delay when loading a buffer
//		highlights new comments
//		takes up less memory

//	comments/suggestions/bug fixes welcome:
//		Morgwn McCarty
//		P.O. 1145
//		Eugene, OR  97440

#include "eel.h"



#define BOOLEAN	int
#define TRUE 1
#define FALSE 0

color_class c_comment_color	=
	{ COLOR(colGREEN,					colBLACK), MCNORM };

color_class cpp_comment_color =
	{ COLOR(colGREEN,					colBLACK), MCNORM };

color_class asm_comment_color =
	{ COLOR(colGREEN,					colBLACK), MCNORM };

color_class string_color =
	{ COLOR(colMAGENTA + colINTENSE,	colBLACK), MCNORM };

color_class preprocessor_color =
	{ COLOR(colRED     + colINTENSE,	colBLACK), MCNORM };

user BOOLEAN colorize_comments = TRUE;	// user flag to colorize buffer

buffer int colorizer_type = 0;			// type of buffer (c or asm)

buffer old_window_start = -1;
buffer old_window_end = -1;

typedef struct regions_tag {
	spot spot1;
	spot spot2;
	int region_num;
} regions_struct;

#define MAX_REGIONS 100
buffer regions_struct regions[MAX_REGIONS + 1];
buffer regions_used = 0;



do_colorizing()
{
	int orig_point;		
	int use_color;		
	int temp;
	int region_idx;						// current region index
	int pos1;							// start of region
	spot spot1;							// temp spot
	spot spot2;							// temp spot
	BOOLEAN first_comment;
	BOOLEAN more_to_search;				// flag for an inner loop
	char *string_search_pattern;
	char *comment_pattern;

	// must be in c or asm file and want to colorize
	if (colorize_comments) {
		if (colorizer_type == 'C') {
			comment_pattern = "[/\"\'%#]";
		} else if (colorizer_type == 'A') {
			comment_pattern = "[;\"\'<]";
		} else {
			return;
		}
	} else {
		return;
	}
	
	// if the endpoints of the window haven't changed, there's nothing to do
	if (window_end == old_window_end) {
		if (window_start == old_window_start) {
			return;
		}
	}

	// save the endpoints for the next time around
	old_window_start = window_start;
	old_window_end = window_end;

	// reset our region index and say we're looking for first comment
	region_idx = 0;
	first_comment = TRUE;
	
	// save cursor location and move to top of window
	orig_point = point;
	point = window_start;

	// scan down through window
	while (point < window_end) {

		// clear our color indicator
		use_color = 0;
		
		// search for our pattern
		if (!re_search(RE_FASTEST, comment_pattern)) {
			break;
		}

		// think we found one!
		pos1 = matchstart;
		switch (character(matchstart)) {

			// maybe it's a comment
			case '/':
				switch (curchar()) {

					// c++ style "//" comment
					case '/':
						nl_forward();
						use_color = color_class cpp_comment_color;
						first_comment = FALSE;
						break;

					// c style "/*" comment
					case '*':

						// search for the end of the comment.  Also, try
						// to figure out if we're in an "open" comment, so
						// we don't highlight the rest of the window.
						for ( ; ; ) {
							if (!search(1, "/")) {
								point = pos1 + 1;
								break;
							}
							if (character(matchstart - 1) == '*') {
								use_color = color_class c_comment_color;
								first_comment = FALSE;
								break;
							} else if (character(matchstart + 1) == '*') {
								point = pos1 + 1;
								break;
							} else if (point >= window_end) {
								point = pos1 + 1;
								break;
							}
						}
						break;

					// just a slash?
					default:

						// check for "*/" near top of screen if we haven't
						// already run into any comments.  This will take
						// care multi-line old-style comments that extend
						// above the top of the window.  This should work
						// good enough in most cases.
						if (first_comment) {
							if (character(matchstart - 1) == '*') {
								pos1 = window_start;
								region_idx = 0; // so this will be first region
								use_color = color_class c_comment_color;
								first_comment = FALSE;
							}
						}
						break;
				}
				break;

			// character constants (or asm strings)
			case '\'':
				if (colorizer_type == 'A') {
					string_search_pattern = "[\'!\n]";
				} else {
					string_search_pattern = "[\'\\\n]";
				}
				goto string_search;

			// strings
			case '\"':
				string_search_pattern = "[\"\\\n]";
string_search:
				// search for the end of the string or character constant.
				// Skip over escaped chars.  Bail out if we hit a newline
				// that's not escaped.
				more_to_search = TRUE;
				while (more_to_search) {
					re_search(RE_FASTEST, string_search_pattern);
					switch(character(matchstart)) {
						case '\'':
						case '\"':
							use_color = color_class string_color;
							more_to_search = FALSE;
							break;
						case '!':
						case '\\':
							point++;
							break;
						case '\n':
						default:
							more_to_search = FALSE;
							break;
					}
				}
				break;
							
			// compiler directive
			case '#':
				nl_forward();
				use_color = color_class preprocessor_color;
				break;

			// asm comment
			case ';':
				nl_forward();
				use_color = color_class asm_comment_color;
				break;

			// nothing found
			default:
				break;
		}

		// highlight the region
		if (use_color) {
			if (region_idx < regions_used) {
				spot1 = regions[region_idx].spot1;
				*spot1 = pos1;
				spot2 = regions[region_idx].spot2;
				*spot2 = point;
				temp = regions[region_idx].region_num;
				modify_region(temp, MRSTART, spot1);
				modify_region(temp, MREND, spot2);
				modify_region(temp, MRCOLOR, use_color);
				region_idx++;
			} else if (region_idx == MAX_REGIONS) {
				point = window_end;		// force loop termination
			} else {					
				spot1 = alloc_spot();
				*spot1 = pos1;
				spot2 = alloc_spot();
				*spot2 = point;
				regions[region_idx].region_num =
					add_region(spot1, spot2, use_color, REGNORM, 0);
				regions[region_idx].spot1 = spot1;
				regions[region_idx].spot2 = spot2;
				regions_used++;
				region_idx++;
			}
		}
	}

	// clear unused regions
	temp = region_idx;
	while (region_idx < regions_used) {
		remove_region(regions[region_idx].region_num);
		free_spot(regions[region_idx].spot1);
		free_spot(regions[region_idx].spot2);
		region_idx++;
	}
	regions_used = temp;					// new top of stack
	
	// restore cursor and refresh screen
	point = orig_point;
	maybe_refresh();
}



new_colorize_when_idle()
{
	do_colorizing();
	colorize_when_idle();
}
REPLACE_FUNC("colorize", "when_idle");



new_colorize_c_mode_hook()
{
	colorizer_type = 'C';
	do_colorizing();
}
REPLACE_FUNC("colorize", "c_mode_hook")



new_colorize_asm_mode_hook()
{
	colorizer_type = 'A';
	tab_size = 8;
	do_colorizing();
}
REPLACE_FUNC("colorize", "asm_mode_hook")



char _asm_mode_name[] = "ASM";

command new_colorize_asm_mode()
{
	major_mode = _asm_mode_name;
	try_calling("asm-mode-hook");
	make_mode();
}
REPLACE_FUNC("colorize", "asm_mode")



new_colorize_suffix_asm()
{
	asm_mode();
}
REPLACE_FUNC("colorize", "suffix_asm")



new_colorize_suffix_inc()
{
	asm_mode();
}
REPLACE_FUNC("colorize", "suffix_inc")



new_colorize_suffix_mac()
{
	asm_mode();
}
REPLACE_FUNC("colorize", "suffix_mac")
